;  TLS.
;
; (c) Indy, 2012
;
; o   (RWL).
;
TLS_ENTRY struct
Teb		PTEB ?
Tid		HANDLE ?	; TID
Gate		PVOID ?
Ip		PVOID ?
TLS_ENTRY ends
PTLS_ENTRY typedef ptr TLS_ENTRY

RDMUX	equ 0
WRMUX	equ 1

%BLOCK macro Mux, Bit
	.repeat
		bts Mux,Bit
	.until !Carry?
endm

%BUNLOCK macro Mux, Bit
	btr Mux, Bit
endm

RWLOCK struct
Mux		DWORD ?
Rc		ULONG ?
RWLOCK ends
PRWLOCK typedef ptr RWLOCK

.data?
TLS_MAX_ENTRIES	equ 800H

g_Tls			TLS_ENTRY TLS_MAX_ENTRIES DUP (<>)
g_TlsLock			RWLOCK <>

.code
; Ecx: PRWLOCK
;
EvReadLock proc C
	assume ecx:PRWLOCK	
	%BLOCK [Ecx].Mux, RDMUX
	inc [ecx].Rc	; mb XADD
	.if [Ecx].Rc == 1
		%BLOCK [Ecx].Mux, WRMUX
	.endif
	%BUNLOCK [Ecx].Mux, RDMUX
	ret
EvReadLock endp

EvReadUnlock proc C
	%BLOCK [Ecx].Mux, RDMUX
	dec [ecx].Rc
	.if Zero?
		%BUNLOCK [Ecx].Mux, WRMUX
	.endif
	%BUNLOCK [Ecx].Mux, RDMUX
	ret
EvReadUnlock endp

EvWriteLock proc C
	%BLOCK RWLOCK.Mux[ecx], WRMUX
	ret
EvWriteLock endp

EvWriteUnlock proc C
	%BUNLOCK RWLOCK.Mux[ecx], WRMUX
	ret
EvWriteUnlock endp

%RLOCK macro pLock
	lea ecx,pLock
	Call EvReadLock
endm

%RUNLOCK macro pLock
	lea ecx,pLock
	Call EvReadUnlock
endm

%WLOCK macro pLock
	lea ecx,pLock
	Call EvWriteLock
endm

%WUNLOCK macro pLock
	lea ecx,pLock
	Call EvWriteUnlock
endm

%ISRLOCK macro Mux
	bt Mux, RDMUX
endm

%ISWLOCK macro Mux
	bt Mux, WRMUX
endm

; +
;     .
;
	assume fs:nothing	
TlsCleaningCycle proc uses esi edi
Local Free:PTLS_ENTRY, MmInfo:MEMORY_BASIC_INFORMATION
	mov eax,TLS_MAX_ENTRIES
	cld
	mov Free,NULL
	lea esi,g_Tls
	assume esi:PTLS_ENTRY
	.repeat
		cmp [esi].Teb,NULL
		mov edi,esi
		.if Zero?
			mov eax,esi
			jmp Exit
		.endif
		invoke ZwQueryVirtualMemory, NtCurrentProcess, [Esi].Teb, MemoryBasicInformation, addr MmInfo, sizeof(MEMORY_BASIC_INFORMATION), NULL
		test eax,eax
		jnz Clear
		cmp MmInfo.State,MEM_COMMIT
		jne Clear
		cmp MmInfo.Protect,PAGE_READWRITE
		mov eax,[esi].Teb
		jne Clear
		cmp MmInfo._Type,MEM_PRIVATE
		mov edx,[esi].Tid
		jne Clear
		.if (TEB.Tib.Self[Eax] != Eax) || (TEB.Cid.UniqueThread[Eax] != Edx)
	Clear:
			.if Free == NULL
				mov Free,edi
			.endif
			mov ecx,sizeof(TLS_ENTRY)/4
			xor eax,eax
			rep stosd
		.endif
		add esi,sizeof(TLS_ENTRY)
		dec eax
	.until Zero?
Load:
	mov eax,Free
Exit:
	ret
TlsCleaningCycle endp

; +
;  TLS    .
;
TlsAdd proc uses esi edi
	mov edx,fs:[TEB.Cid.UniqueThread]
	%WLOCK g_TlsLock
	mov ecx,fs:[TEB.Tib.Self]
	lea esi,g_Tls
	assume esi:PTLS_ENTRY
	mov eax,TLS_MAX_ENTRIES
	.repeat
	   cmp [esi].Teb,NULL
	   je Alloc
	   cmp [esi].Teb,ecx
	   je IsTid
   Scan:
	   add esi,sizeof(TLS_ENTRY)
	   dec eax
	.until Zero?
	;   .   .
	invoke TlsCleaningCycle	;   .
	jmp Exit
IsTid:
	cmp [esi].Tid,edx
	jne Scan
	jmp Load
;   ,  .
Alloc:
	mov [esi].Teb,ecx
	mov [esi].Tid,edx
Load:
	mov eax,esi
Exit:
	%WUNLOCK g_TlsLock
	test eax,eax
	ret
TlsAdd endp

; +
;  TLS  .
;
TlsGet proc uses esi
	mov edx,fs:[TEB.Cid.UniqueThread]
	%RLOCK g_TlsLock
	mov ecx,fs:[TEB.Tib.Self]
	lea esi,g_Tls
	assume esi:PTLS_ENTRY
	mov eax,TLS_MAX_ENTRIES
	.repeat
	   cmp [esi].Teb,NULL
	   je Error
	   cmp [esi].Teb,ecx
	   je IsTid
   Scan:
	   add esi,sizeof(TLS_ENTRY)
	   dec eax
	.until Zero?
Error:
	xor eax,eax
Exit:
	%RUNLOCK g_TlsLock
	test eax,eax
	ret
IsTid:
	cmp [esi].Tid,edx
	jne Scan
Load:
	mov eax,esi
	jmp Exit
TlsGet endp